#![no_std]

mod util;

use core::{cell::RefCell, fmt::Debug, iter::repeat};

use alloc::{
	rc::Rc,
	string::{String, ToString},
	vec::Vec,
};

use ast::{Block, FunDef, Short};
use chumsky::{error::Simple, Parser};
use hashbrown::HashMap;

extern crate alloc;

#[derive(Debug)]
pub struct Context {
	pub frame: Rc<RefCell<Scope>>,
}

#[derive(Debug)]
pub enum Error {
	ParsingErr(Vec<Simple<char>>),
	EvalErr(EvalError),
}

#[derive(Debug)]
pub enum EvalError {
	Exit,
	NotFound(String),
	IllegalShort(Short),
	WrongType, // add type names ?
}

impl Context {
	pub fn empty() -> Self {
		let frame = Rc::new(RefCell::new(Scope::new(None)));
		Self { frame }
	}

	pub fn eval(&mut self, script: String) -> Result<(Block, Value), Error> {
		let parsed = match Block::program_parser().parse(script) {
			Ok(parsed) => parsed,
			Err(err) => Err(Error::ParsingErr(err))?,
		};
		let value = match parsed.eval(self) {
			Ok((_short, value)) => value,
			Err(err) => Err(Error::EvalErr(err))?,
		};
		Ok((parsed, value))
	}

	pub fn define(&mut self, name: impl ToString, value: Value) {
		self.scope().borrow_mut().define(name.to_string(), value)
	}

	pub fn define_built(&mut self, name: impl ToString, op: impl Fn(Vec<Value>, &mut Context) -> FunRes + 'static) {
		let def = FunImpl::Built(Rc::new(op) as Rc<BuiltFun>);
		let function = Function {
			capturing: Rc::new(RefCell::new(Scope::new(None))),
			def,
		};
		self.define(name.to_string(), Value::Function(Rc::new(RefCell::new(function))))
	}

	pub fn scope(&self) -> Rc<RefCell<Scope>> {
		self.frame.clone()
	}
}

#[derive(Debug, Clone)]
pub struct Scope {
	pub parent: Option<Rc<RefCell<Self>>>,
	pub defs: HashMap<String, Value>,
}

impl Scope {
	pub fn new(parent: Option<Rc<RefCell<Self>>>) -> Self {
		let defs = HashMap::new();
		Self { defs, parent }
	}

	pub fn child<F: FnOnce(&mut Context) -> O, O>(context: &mut Context, op: F) -> O {
		let parent = context.frame.clone();
		let child = Rc::new(RefCell::new(Scope::new(Some(parent.clone()))));
		context.frame = child;
		let result = op(context);
		context.frame = parent;
		result
	}

	pub fn alt<F: FnOnce(&mut Context) -> O, O>(context: &mut Context, captures: Rc<RefCell<Self>>, op: F) -> O {
		let parent = context.frame.clone();
		let alt = Rc::new(RefCell::new(Scope::new(Some(captures.clone()))));
		context.frame = alt;
		let result = op(context);
		context.frame = parent;
		result
	}

	pub fn define(&mut self, name: String, value: Value) {
		self.defs.insert(name, value);
	}

	pub fn update(&mut self, name: &str, value: Value) -> Result<(), EvalError> {
		let Some(variable) = self.defs.get_mut(name) else {
			if let Some(parent) = &self.parent {
				return parent.borrow_mut().update(name, value);
			}
			return Err(EvalError::NotFound(name.into()));
		};
		*variable = value;
		Ok(())
	}

	pub fn get(&self, name: &str) -> Result<Value, EvalError> {
		let Some(value) = self.defs.get(name) else {
			if let Some(parent) = &self.parent {
				return parent.borrow_mut().get(name);
			}
			return Err(EvalError::NotFound(name.into()));
		};
		Ok(value.clone())
	}
}

#[derive(Debug, Clone)]
pub enum Value {
	None,
	Bool(bool),
	Num(f64),
	Str(String),
	Object(Rc<RefCell<HashMap<String, Value>>>),
	Array(Rc<RefCell<Vec<Value>>>),
	Function(Rc<RefCell<Function>>),
}

impl Value {
	pub fn serialize(&self) -> String {
		match self {
			Self::None => "none".to_string(),
			Self::Bool(b) => b.to_string(),
			Self::Num(n) => n.to_string(),
			Self::Str(s) => s.to_string(),
			Self::Object(o) => {
				let mut res = "{".to_string();
				let mut first = true;
				for (name, value) in o.borrow().iter() {
					if first {
						res += " ";
						res += name;
						res += ": ";
						res += &value.serialize();
						first = false;
					} else {
						res += ", ";
						res += name;
						res += ": ";
						res += &value.serialize();
					}
				}
				res += " }";
				res
			}
			Self::Array(a) => {
				let mut res = "[".to_string();
				let mut first = true;
				for value in a.borrow().iter() {
					if first {
						res += " ";
						res += &value.serialize();
						first = false;
					} else {
						res += ", ";
						res += &value.serialize();
					}
				}
				res += " ]";
				res
			}
			Self::Function(_f) => "<function>".to_string(),
		}
	}
}

pub enum InstrOutput {
	Short(Short),
	Value(Value),
}

#[derive(Debug, Clone)]
pub struct Function {
	pub capturing: Rc<RefCell<Scope>>,
	pub def: FunImpl,
}

#[derive(Clone)]
pub enum FunImpl {
	Def(FunDef),
	Built(Rc<BuiltFun>),
}

pub type FunRes = Result<Value, EvalError>;
pub type BuiltFun = dyn Fn(Vec<Value>, &mut Context) -> FunRes;

impl Debug for FunImpl {
	fn fmt(&self, f: &mut core::fmt::Formatter<'_>) -> core::fmt::Result {
		match self {
			Self::Built(_) => f.debug_struct("Built").finish(),
			Self::Def(def) => f.debug_tuple("Def").field(def).finish(),
		}
	}
}

impl Function {
	pub fn call(&self, args: Vec<Value>, context: &mut Context) -> Result<Value, EvalError> {
		match &self.def {
			FunImpl::Built(op) => op(args, context),
			FunImpl::Def(def) => Scope::alt(context, self.capturing.clone(), |context| {
				let names = def.args.iter().cloned();
				for (name, value) in names.zip(args.into_iter().chain(repeat(Value::None))) {
					context.scope().borrow_mut().define(name, value);
				}
				match def.body.eval(context) {
					Err(err) => Err(err),
					Ok((None, value)) => Ok(value),
					Ok((Some(Short::Return(ret)), _)) => Ok(ret),
					Ok((Some(s), _)) => Err(EvalError::IllegalShort(s)),
				}
			}),
		}
	}
}

pub mod ast;
